(* main.ml *)

open Common
open Printf
open Lexing

let lex = Lexing.from_channel stdin
let history = History.mk 40

(* Affiche la position dans l'entrée d'un symbole non reconnu par le lexer *)
let display_error l msg x =
	let pos = l.lex_start_p in	
	let line = pos.pos_lnum in
	let bpos = pos.pos_cnum - pos.pos_bol in
	let epos = bpos + lexeme_end l - lexeme_start l in
	if epos=bpos+1 then 
		eprintf ("Ligne %d, caractère %d: %s: `%s'.\n%!") line bpos msg x
	else
		eprintf ("Ligne %d, caractères %d-%d: %s: `%s'.\n%!") line bpos epos msg x;
		Lexing.flush_input l; 
		Lexing.new_line l

let rec next_command l =
	try Parser.main Lexer.token l with 
		| Lexer.Error c -> 
			display_error l "unrecognised token starting with" ( String.make 1 c );
			printf "> %!";
			next_command l
		| Parsing.Parse_error -> 
			display_error l "unexpected token" ( lexeme l );
			printf "> %!";
			next_command l
		| Command_error s ->
			eprintf "Tactique non définie : %s\n%!" s;
			printf "> %!";
			next_command l
			
(* Tant que l'on traite un fichier le programme est silencieux et n'enregistre
	rien dans l'historique *)
let rec file_loop l state =
		try 
			let ( command, _, _ ) = next_command l in
			file_loop l ( process_command ( state, command ) )
		with End_of_file -> state

and process_command = function
	| state, Load f ->
		begin
		try
			let file = open_in f in
			let l = Lexing.from_channel file in
			let res = file_loop l state in
			close_in file;
			res
		with Sys_error s -> eprintf "Erreur de lecture : %s\n%!" s; state
		end
	| state, Save f -> 
		begin
		try
			let file = open_out f in
			History.print file history;
			close_out file;
		with Sys_error s -> eprintf "Erreur d'écriture : %s\n%!" s
		end;
		state
	| state, Undo -> ( 
		try History.undo history 
		with Not_found -> printf "Historique vide\n%!"; state )
	| _, Start seq -> Some ( State.start seq )
	| Some state, Apply tac -> 
		begin
		try Some ( State.apply tac state )
		with State.Invalid_tactic s ->
			eprintf "Erreur dans la tactique: %s\n%!" s;
			Some state
		end
	| Some state, Focus i -> Some ( State.focus i state )
	| Some state, Finish ->
		begin
		try
			let proof = State.finish state in
			printf "Preuve terminée :\n%a%!"
			Proof.print_proof proof;
			( if Proof.is_intuit proof then printf "Preuve intuitioniste\n%!" );
			None
		with 
		| Proof.Invalid_proof s ->
			eprintf "Preuve mal formée : %s\n%!" s;
			Some state
		| State.Invalid_tactic s ->
			eprintf "Erreur : %s\n%!" s;
			Some state
		end
	| None, Apply _
	| None, Focus _
	| None, Finish ->
		eprintf "Erreur : Veuillez définir un but avec /start\n%!"; None
	| _, Exit -> exit 0

let record state ( start, len ) =
	let str = String.sub lex.lex_buffer start len in
	History.add history ( str, state )

let process state ( commande, start, len ) =
	record state ( start, len );
	process_command ( state, commande )

let rec main_loop state =
	begin
	match state with
		| None -> 
			printf "Pas de but défini\n> %!";
		| Some s -> 
			let l = State.nb_buts s in
			if l == 0 then
				printf "Aucun but restant, essayez /qed\n> %!"
			else
				printf "But 1/%d: %a\n> %!"
				l Sequent.print_sequent ( State.but s )
	end;
	try main_loop ( process state ( next_command lex ) ) 
	with End_of_file -> exit 0

let _ = 
	main_loop None 

