(* proof.ml *)

open Formule
open Sequent

(* représentation arborescente de la preuve
	sous la forme (prémisses,conclusion,règle) *)
type sym_proof = Node of sym_proof list * sequent * string
(* intuitioniste?, premisses, arbre de la preuve *)
type info = bool * ( sequent list ) * sym_proof

type proof = sequent * info

exception Invalid_proof of string

let failwith s = raise( Invalid_proof s )

(* Fusionne deux listes de prémisses, trie et enlève les doublons *)
let merge l1 l2 =
	let l = List.sort compare ( List.rev_append l1 l2 ) in
	let rec clean accu = function
		| None -> ( function
			| [] -> accu
			| t::q -> clean ( t :: accu ) ( Some t ) q )
		| Some x -> ( function
			| [] -> accu
			| t :: q when x = t -> clean accu ( Some t ) q
			| t :: q -> clean ( t :: accu ) ( Some t ) q ) in
	clean [] None l

(* Synthétise la partie info de la preuve pour les règles à deux prémisses *)
let combine s seq ( i1, p1, sp1 ) ( i2, p2, sp2 ) =
	( i1 && i2, merge p1 p2, Node ( [ sp1; sp2 ], seq, s) )

(* Synthétise la partie info de la preuve pour les règles à une prémisse *)
let add_info b s seq ( i, p, sp ) = ( i && b, p, Node ( [ sp ], seq, s ) )

(* Implémentation des règles *)

let assume seq = ( seq, ( true, [ seq ], Node ( [], seq, "Prémisse" ) ) )

let axiom ( h, c ) =
	if List.exists ( function ( _, f ) -> f = c ) h then
		( ( h, c ), ( true, [], Node ( [], ( h, c ), "Axiom" ) ) )
	else failwith "erreur" 

let aff i_a a ( ( h, c ), i ) =
	let seq = ( ( i_a, a ) :: h, c ) in
	( seq, add_info true "Aff" seq i )

let and_r ( ( h_a, a ), ia ) ( ( h_b, b ), ib ) = 
	if eq_hyp h_a h_b then
		let seq = ( h_a, And ( a, b ) ) in
		( seq, combine "And_r" seq ia ib )
	else failwith "erreur"

let and_l i_ab i_a i_b ( ( h, c ), i ) = 
	try
		let a = List.assoc i_a h
		and b = List.assoc i_b h 
		and h_ab = List.remove_assoc i_a ( List.remove_assoc i_b h ) in
		let seq = ( ( i_ab, And ( a, b ) ) :: h_ab , c ) in
		( seq, add_info true "And_l" seq i )
	with
	| Not_found -> failwith "erreur"

let or_l i_ab i_a i_b ( ( h_a, ca ), ia ) ( ( h_b, cb ), ib ) =
	try
		let a = List.assoc i_a h_a
		and b = List.assoc i_b h_b
		and h_ab1 = List.remove_assoc i_a h_a 
		and h_ab2 = List.remove_assoc i_b h_b in
		if eq_hyp h_ab1 h_ab2 && ca = cb then
			let seq = ( ( i_ab, Or ( a, b ) ) :: h_ab1 , ca ) in
			( seq, combine "Or_l" seq ia ib )
		else failwith "erreur"
	with
	| Not_found -> failwith "erreur"
		
let or_r1 b ( ( h, a ), i ) = 
	let seq = ( h, Or ( a, b ) ) in
	( seq, add_info true "Or_r1" seq i )

let or_r2 a ( ( h, b ), i ) =
	let seq = ( h, Or ( a, b ) ) in
	( seq, add_info true "Or_r2" seq i )
	
let imply_l i_ab i_b ( ( h, a ), ia ) ( ( hb, c ), ib ) =
	try
		let b = List.assoc i_b hb in
		if eq_hyp ( ( i_b, b ) :: h ) hb then
			let seq = ( ( i_ab, Imply ( a, b ) ) :: h, c ) in
			( seq, combine "Imply_l" seq ia ib )
		else failwith "erreur"
	with
	| Not_found -> failwith "erreur"

let imply_r i_a ( ( ha, b ), i ) =
	try
		let a = List.assoc i_a ha
		and h = List.remove_assoc i_a ha in
		let seq = ( h, Imply ( a, b ) ) in
		( seq, add_info true "Imply_r" seq i )	
	with
	| Not_found -> failwith "erreur"

let neg_l i_na ( ( h, a ), i ) = 
	let seq = ( ( i_na, Not a ) :: h, Bottom ) in
	( seq, add_info true "Neg_l" seq i )

let neg_r i_a ( ( ha, c ), i ) =
	try
		let a = List.assoc i_a ha
		and h = List.remove_assoc i_a ha in
		if c = Bottom then 
			let seq = ( h, Not a ) in
			( seq, add_info true "Neg_r" seq i ) else failwith "erreur"
	with
	| Not_found -> failwith "erreur"

let cut i ( ( ha, ca ), ia ) ( ( hb, cb ), ib ) =
	if ( eq_hyp ( ( i, ca ) :: ha ) hb ) then
		let seq = ( ha, cb ) in
		( seq, combine "Cut" seq ia ib ) 
	else failwith "erreur"

let abs i ( ( ha, ca ), ia ) =
	try
		let na = List.assoc i ha
		and f = function
			| Not x -> x
			| _ -> raise Not_found in
		let a = f na
		and h = List.remove_assoc i ha in
		if ca = Bottom then 
			let seq = ( h, a ) in
			( seq, add_info true "Abs" seq ia ) else failwith "erreur"
	with
	| Not_found -> failwith "erreur"

(* Fonctions d'affichage *)

let is_intuit ( _, ( i, _, _ ) ) = i

let rec indent o = function
	| 0 -> ()
	| i when i>0 -> Printf.fprintf o "|   "; indent o (i-1)
	| _ -> assert false

(* Affiche la preuve en notation linéaire avec indentation *)
let rec print_sym_proof i o s_proof =
	let print_iter f o = List.iter ( f o ) in
	let ( sup, inf, rule ) = match s_proof with
		| Node ( [], seq, s ) -> ( [], seq, s )
		| Node ( l, seq, s ) -> ( l, seq, s ) in
	Printf.fprintf o "%a>%a  %s\n%a"
		indent i
		print_sequent inf
		rule
		( print_iter ( print_sym_proof (i+1) ) ) sup

let print_proof1 o ( _, ( _, _, s_proof ) ) = print_sym_proof 0 o s_proof

(* Exporte la preuve pour être insérée dans
	un document Latex avec le module mathpartir *)
let rec latex_sym_proof o s_proof =
	let to_tex = function
		| "Prémisse" -> "Hyp"
		| "Axiom" -> "Ax"
		| "Aff" -> "Aff"
		| "And_l" -> "$\\wedge_{\\text{g}}$"
		| "And_r" -> "$\\wedge_{\\text{d}}$"
		| "Or_l" -> "$\\vee_{\\text{g}}$"
		| "Or_r1" -> "$\\vee_{\\text{d1}}$"
		| "Or_r2" -> "$\\vee_{\\text{d2}}$"
		| "Imply_l" -> "$\\rightarrow_{\\text{g}}$"
		| "Imply_r" -> "$\\rightarrow_{\\text{d}}$"
		| "Neg_l" -> "$\\neg_{\\text{g}}$"
		| "Neg_r" -> "$\\neg_{\\text{d}}$"
		| "Abs" -> "$\\bot_{\\text{c}}$"
		| "Cut" -> "Cut"
		| s -> s in
	let rec print_iter f o = function
		| [] -> Printf.fprintf o " "
		| [ a ] -> f o a
		| t :: q -> Printf.fprintf o "%a \\\\ %a" f t ( print_iter f ) q in
	let ( sup, inf, rule ) = match s_proof with Node ( l, seq, s ) -> ( l, seq, s ) in
	Printf.fprintf o "\\inferrule*[Right=%s]{ %a }{ %a }"
		( to_tex rule )
		( print_iter latex_sym_proof ) sup
		latex_print_sequent inf

let latex_print_proof o ( _, ( _, _, s_proof ) ) = 
	Printf.fprintf o "\\begin{mathpar}\n%a\n\\end{mathpar}" latex_sym_proof s_proof
		 
let print_premisses o ( _, ( _, l, _ ) ) =
	if List.length l <> 0 then
		Printf.fprintf o "Prémisses:\n";
		let rec f = function
			| [] -> ()
			| t :: q -> Printf.fprintf o "%a\n" print_sequent t; f q in
		f l

let print_proof = match Sys.argv with
	[|_;"-latex"|] -> latex_print_proof
	| _ -> print_proof1

let print_result o ( p, _ ) = print_sequent o p
