(* state.ml *)

open Proof
open Sequent
open Formule

type state = sequent list * ( proof list -> proof )
type tactic = sequent -> state

(*si une tactique est appliquée à un état incompatible*)
exception Invalid_tactic of string 

let failwith s = raise( Invalid_tactic s )

let nb_buts ( buts, _ ) = List.length buts

let but = function ( t :: _, _ ) -> t | ( [], _ ) -> raise Not_found

(* Commandes *)

let start s =
	( [ s ], function [ p ] -> p | _ -> failwith "start" )

let finish = function 
	| ( [], f ) ->
		begin
		try f( [] )
		with Invalid_proof s ->
			failwith ( "Erreur de construction :" ^ s )
		end
	| _ -> failwith "Preuve non terminée"

let apply tac = function
	| ( [], _ ) -> failwith "Pas de séquent cible"
	| ( t :: q, f ) ->
		let ( sequents, g ) = tac t in
		let rec split = function
			| ( i, a, b ) when i = 0 -> ( a, List.rev b )
			| ( i, t :: q, b ) -> split( i - 1, q, t :: b )
			| _ -> raise ( Invalid_proof "Apply" ) in 
		let f p =
			let ( a, b ) = split( List.length( sequents ), p , [] ) in
			f( g( b ) :: a ) in
		( sequents @ q, f )

let focus i ( sequents, f ) =
	if i >= 1 && i <= List.length sequents then  
		let rec pull j accu = function
			| t :: q when j > 1 -> pull ( j - 1 ) ( t :: accu ) q
			| t :: q when j = 1 -> t :: ( List.rev_append accu q )
			| _ -> assert false in
		let g =
			let rec push x j accu = function
				| t :: q when j > 1 -> push x ( j - 1 ) ( t :: accu ) q
				| l when j = 1 -> List.rev_append accu ( x :: l )
				| _ -> raise ( Invalid_proof "Focus" ) in
			( function
				| [] -> f []
				| ( t :: q )  -> f ( push t i [] q ) ) in
		( pull i [] sequents, g )
	else if i = 1 then
		( sequents, f )
	else failwith ( "Indice non valide : "^string_of_int(i) )

(* Tactiques élémentaires *)

let tac_assume seq = 
	let f = function
		| [] -> assume seq
		| _ ->  raise ( Invalid_proof "Tactique Assume" ) in
	( [], f )

let tac_axiom ( h, c ) =
	if List.exists ( function ( _, f ) -> f = c ) h then
		let f = function
			| [] -> axiom ( h, c )
			| _ ->  raise ( Invalid_proof "Tactique Axiom" ) in
		( [], f )
	else failwith "Pas un axiome"

let tac_aff i_a ( h, c ) =
	try
		let a = List.assoc i_a h
		and h_a = List.remove_assoc i_a h in
		let f = function
			| [ p ] -> aff i_a a p
			| _ ->  raise ( Invalid_proof "Tactique Aff" ) in
		( [ ( h_a, c ) ], f )
	with
	| Not_found -> failwith "hypothèse non trouvée"

let tac_and_l i_ab i_a i_b ( h, c ) = 
	let extract = function 
		| And ( a, b ) -> ( a, b )
		| _ -> raise Not_found in
	try 
		let ( a, b ) = extract ( List.assoc i_ab h ) in
		let h_a_b = ( i_a, a ) :: ( i_b, b ) :: ( List.remove_assoc i_ab h )
		and f = function 
			| [ p ] -> and_l i_ab i_a i_b p
			| _ -> raise ( Invalid_proof "Tactique And_l" ) in
		( [ ( h_a_b, c ) ], f ) 
	with
	| Not_found -> failwith "/\\ non trouvé"
	
let tac_and_r = function
	| ( h, And ( a, b ) ) ->
		let f = function
		| [ pa; pb ] -> ( and_r pa pb ) 
		| _ -> raise ( Invalid_proof "Tactique And_r" ) in
		( [ ( h, a ); ( h, b ) ], f )
	| _ -> failwith "Pas un /\\"

let tac_or_l i_ab i_a i_b ( h, c ) =
	let extract = function 
		| Or ( a, b ) -> ( a, b )
		| _ -> raise Not_found in
	try 
		let ( a, b ) = extract ( List.assoc i_ab h )
		and h = List.remove_assoc i_ab h in
		let h_a = ( i_a, a ) :: h
		and h_b = ( i_b, b ) :: h
		and f = function 
			| [ pa; pb ] -> or_l i_ab i_a i_b pa pb
			| _ -> raise ( Invalid_proof "Tactique Or_l" ) in
		( [ ( h_a, c ); ( h_b, c ) ], f ) 
	with
	| Not_found -> failwith "\\/ non trouvé"

let tac_or_r1 = function
	| ( h, Or ( a, b ) ) ->
		let f = function
		| [ p ] -> or_r1 b p  
		| _ -> raise ( Invalid_proof "Tactique Or_r1" ) in
		( [ ( h, a ) ], f )
	| _ -> failwith "Pas un \\/"
	
let tac_or_r2 = function
	| ( h, Or ( a, b ) ) ->
		let f = function
		| [ p ] -> or_r2 a p  
		| _ -> raise ( Invalid_proof "Tactique Or_r2" ) in
		( [ ( h, b ) ], f )
	| _ -> failwith "Pas un \\/"

let tac_imply_l i_ab i_b ( h, c ) =
	let extract = function 
		| Imply ( a, b ) -> ( a, b )
		| _ -> raise Not_found in
	try 
		let ( a, b ) = extract ( List.assoc i_ab h )
		and h = List.remove_assoc i_ab h in
		let hb = ( i_b, b ) :: h
		and f = function 
			| [ pa; pc ] -> imply_l i_ab i_b pa pc
			| _ -> raise ( Invalid_proof "Tactique Imply_l" ) in
		( [ ( h, a ); ( hb, c ) ], f )
	with
	| Not_found -> failwith "-> non trouvé"

let tac_imply_r i_a = function
	| ( h, Imply ( a, b ) ) ->
		let f = function
		| [ p ] -> imply_r i_a p  
		| _ -> raise ( Invalid_proof "Tactique Imply_r" ) in
		( [ ( ( i_a, a ) :: h, b ) ], f )
	| _ -> failwith "Pas un ->"
	
let tac_neg_l i_na ( h, c ) =
	let extract = function 
		| Not x -> x
		| _ -> raise Not_found in
	try 
		let a = extract ( List.assoc i_na h )
		and h = List.remove_assoc i_na h
		and f = function 
			| [ p ] -> neg_l i_na p
			| _ -> raise ( Invalid_proof "Tactique Neg_l" ) in
		if c = Bottom then ( [ ( h, a ) ], f ) else raise Not_found 
	with
	| Not_found -> failwith "! non trouvé"

let tac_neg_r i_a = function
	| ( h, Not a ) ->
		let f = function
		| [ p ] -> neg_r i_a p  
		| _ -> raise ( Invalid_proof "Tactique Neg_r" ) in
		( [ ( ( i_a, a ) :: h, Bottom ) ], f )
	| _ -> failwith "Pas un ->"
	
let tac_cut i a ( h, c ) =
	let f = function
		| [ pa; pb ] -> ( cut i pa pb )
		| _ -> raise ( Invalid_proof "Tactique Cut" ) in
	( [ ( h, a ); ( ( i, a ) :: h, c ) ], f )

let tac_abs i ( h, c ) =
	let f = function
		| [ p ] -> abs i p
		| _ -> raise ( Invalid_proof "Tactique Abs" ) in
	( [ ( i, Not c ) :: h, Bottom ], f )

let tac_skip seq = 
	let f = function
		| [ p ] -> p
		| _ -> raise ( Invalid_proof "Tactique Skip" ) in
	( [ seq ], f )

(* Composition de tactiques *)
let tac_comp ( f : tactic ) ( g : tactic ) seq =
	let ( sequents, p ) = f seq in
	let ( seqlist, f_list ) = List.split ( List.map g sequents ) in
	let lglist = List.map List.length seqlist in
	let rec split = function
		| [], [], [], [], res -> List.rev res
		| ( i :: qi ), accu_proof, ( p :: qp ), f_list, res when i > 0 ->
			split ( ( ( i - 1 ) :: qi ), ( p :: accu_proof ), qp, f_list, res )
		| ( 0 :: qi ), accu_proof, proofs, t :: q, res ->
			split ( qi, [], proofs, q, ( t ( List.rev accu_proof ) ) :: res )
		| _ -> assert false in
	let h proof_list = p ( split ( lglist, [], proof_list, f_list, [] ) ) in
	( List.flatten seqlist, h )	

(* Alternative *)
let tac_cond f g seq =
	try f seq 
	with
	| Invalid_tactic _ -> g seq

(* Répétition *)
let rec tac_repeat f seq = tac_cond ( tac_comp f ( tac_repeat f ) ) tac_skip seq

